==> cx88-1.4.1.tar.gz: cx88-1.4.1/client/dvb/cx88_dvb_buffer.cpp <==
/*-
 * Copyright 2011 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
extern "C"
{
   #include <cuse4bsd.h>
}
#include "cx88_dvb_buffer.h"

cx88_dvb_buffer::cx88_dvb_buffer(size_t size, int &error)
   : m_buffer(NULL),
     m_size(size),
     m_read_pos(0),
     m_write_pos(0),
     m_break_pos(m_size),
     m_state(CX88_DVB_BUF_UNINITIALIZED),
     m_timeout(false),
     m_write_rollover(false),
     m_canceled(false)
{
   if (error)
   {
      return;
   }
   m_buffer = new uint8_t[size];
   if (m_buffer == NULL)
   {
      error = CUSE_ERR_NO_MEMORY;
      return;
   }
   error = pthread_mutex_init(&m_mutex, NULL);
   if (!error)
   {
      m_state = CX88_DVB_BUF_MUTEX_CREATED;
      error = pthread_mutex_init(&m_read_lock, NULL);
   }
   if (!error)
   {
      m_state = CX88_DVB_BUF_READ_LOCK_CREATED;
      error = pthread_cond_init(&m_data_avail, NULL);
   }
   if (!error)
   {
      m_state = CX88_DVB_BUF_CV_CREATED;
   }
}

cx88_dvb_buffer::~cx88_dvb_buffer(void)
{
   delete[] m_buffer;
   m_buffer = NULL;
   m_size = 0;
   m_read_pos = 0;
   m_write_pos = 0;
   if (m_state >= CX88_DVB_BUF_CV_CREATED)
   {
      pthread_cond_destroy(&m_data_avail);
   }
   if (m_state >= CX88_DVB_BUF_READ_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_read_lock);
   }
   if (m_state >= CX88_DVB_BUF_MUTEX_CREATED)
   {
      pthread_mutex_destroy(&m_mutex);
   }
   m_state = CX88_DVB_BUF_UNINITIALIZED;
}

int cx88_dvb_buffer::read(void *user_ptr, int fflags, size_t &size)
{
   int error = 0;
   size_t bytes_remaining = size;
   pthread_mutex_lock(&m_read_lock);
   pthread_mutex_lock(&m_mutex);
   m_canceled = false;
   while (bytes_remaining > 0)
   {
      if (!m_write_rollover && (m_write_pos == m_read_pos))
      {
         if (fflags & CUSE_FFLAG_NONBLOCK)
         {
            break;
         }
         if (!m_timeout)
         {
            pthread_cond_wait(&m_data_avail, &m_mutex);
         }
         else
         {
            if (pthread_cond_timedwait(&m_data_avail, &m_mutex, &m_deadline) != 0)
            {
               error = CUSE_ERR_OTHER;
               break;
            }
         }
      }
      if (m_canceled)
      {
         error = CUSE_ERR_SIGNAL;
         break;
      }
      if (!m_write_rollover && (m_write_pos == m_read_pos))
      {
         if (cuse_got_peer_signal() == 0)
         {
            error = CUSE_ERR_SIGNAL;
            break;
         }
         else
         {
            continue;
         }
      }
      size_t write_pos = ((m_break_pos < m_size) ? m_break_pos : m_write_pos);
      size_t bytes_to_copy = write_pos - m_read_pos;
      if (bytes_to_copy == 0)
      {
         bytes_to_copy = m_size;
      }
      if (m_read_pos > write_pos)
      {
         bytes_to_copy = m_size - m_read_pos + write_pos;
      }
      if (bytes_to_copy > bytes_remaining)
      {
         bytes_to_copy = bytes_remaining;
      }
      size_t remainder = 0;
      if ((m_read_pos + bytes_to_copy) > m_size)
      {
         remainder = m_read_pos + bytes_to_copy - m_size;
      }
      uint8_t *read_pos = m_buffer + m_read_pos;
      pthread_mutex_unlock(&m_mutex);
      error = cuse_copy_out(read_pos, (uint8_t*)user_ptr + size - bytes_remaining, bytes_to_copy - remainder);
      bytes_remaining -= (bytes_to_copy - remainder);
      if (!error && (remainder > 0))
      {
         error = cuse_copy_out(m_buffer, (uint8_t*)user_ptr + size - bytes_remaining, remainder);
         bytes_remaining -= remainder;
      }
      pthread_mutex_lock(&m_mutex);
      if (error)
      {
         break;
      }
      m_write_rollover = false;
      m_read_pos += bytes_to_copy;
      if (m_read_pos >= m_size)
      {
         m_read_pos -= m_size;
      }
      if (m_read_pos == m_break_pos)
      {
         m_break_pos = m_size;
         break;
      }
   }
   pthread_mutex_unlock(&m_mutex);
   pthread_mutex_unlock(&m_read_lock);
   size = size - bytes_remaining;
   return error;
}

int cx88_dvb_buffer::write(const void *data, size_t size)
{
   pthread_mutex_lock(&m_mutex);
   size_t space_avail = m_read_pos - m_write_pos;
   if (m_write_pos >= m_read_pos)
   {
      space_avail = m_size - m_write_pos + m_read_pos;
   }
   if (m_write_rollover || (space_avail < size))
   {
      pthread_mutex_unlock(&m_mutex);
      return CUSE_ERR_OTHER;
   }
   size_t remainder = 0;
   if ((m_write_pos + size) > m_size)
   {
      remainder = m_write_pos + size - m_size;
   }
   uint8_t *write_pos = m_buffer + m_write_pos;
   pthread_mutex_unlock(&m_mutex);
   memcpy(write_pos, data, size - remainder);
   if (remainder > 0)
   {
      memcpy(m_buffer, (uint8_t*)data + size - remainder, remainder);
   }
   pthread_mutex_lock(&m_mutex);
   m_write_pos += size;
   if (m_write_pos >= m_size)
   {
      m_write_pos -= m_size;
   }
   m_write_rollover = (m_write_pos == m_read_pos);
   pthread_cond_signal(&m_data_avail);
   cuse_poll_wakeup();
   pthread_mutex_unlock(&m_mutex);
   return 0; 
}

void cx88_dvb_buffer::reset_pointers(void)
{
   m_read_pos = 0;
   m_write_pos = 0;
   m_write_rollover = false;
   m_break_pos = m_size;
}

int cx88_dvb_buffer::resize(size_t newsize)
{
   if (newsize == m_size)
   {
      return 0;
   }
   int error = 0;
   pthread_mutex_lock(&m_mutex);
   if (m_write_rollover || (m_read_pos != m_write_pos))
   {
      error = CUSE_ERR_BUSY;
   }
   else
   {
      uint8_t *newbuffer = new uint8_t[newsize];
      if (newbuffer == NULL)
      {
         error = CUSE_ERR_NO_MEMORY;
      }
      else
      {
         delete[] m_buffer;
         m_buffer = newbuffer;
         m_size = newsize;
         reset_pointers();
      }
   }
   pthread_mutex_unlock(&m_mutex);
   return error;
}

int cx88_dvb_buffer::poll(int events)
{
   int mask = 0;
   pthread_mutex_lock(&m_mutex);
   if (m_write_rollover || (m_read_pos != m_write_pos))
   {
      mask = events & CUSE_POLL_READ;
   }
   pthread_mutex_unlock(&m_mutex);
   return mask;
}

void cx88_dvb_buffer::reset(void)
{
   pthread_mutex_lock(&m_mutex);
   reset_pointers();
   pthread_mutex_unlock(&m_mutex);
}

void cx88_dvb_buffer::set_timeout(unsigned long timeout_s)
{
   pthread_mutex_lock(&m_mutex);
   if (timeout_s)
   {
      m_timeout = true;
      clock_gettime(CLOCK_MONOTONIC, &m_deadline);
      m_deadline.tv_sec += timeout_s;
   }
   else
   {
      m_timeout = false;
   }   
   pthread_mutex_unlock(&m_mutex);
}

void cx88_dvb_buffer::cancel_transfer(void)
{
   pthread_mutex_lock(&m_mutex);
   m_canceled = true;
   pthread_cond_signal(&m_data_avail);
   pthread_mutex_unlock(&m_mutex);
}

==> cx88-1.4.8.tar.gz: cx88-1.4.8/client/dvb/cx88_dvb_buffer.cpp <==
/*-
 * Copyright 2011 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <new>
extern "C"
{
   #include <cuse4bsd.h>
}
#include "cx88_dvb_buffer.h"

using namespace std;

cx88_dvb_buffer::cx88_dvb_buffer(size_t size, int &error)
   : m_buffer(NULL),
     m_size(size),
     m_read_pos(0),
     m_write_pos(0),
     m_break_pos(m_size),
     m_state(CX88_DVB_BUF_UNINITIALIZED),
     m_timeout(false),
     m_write_rollover(false),
     m_canceled(false)
{
   if (error)
   {
      return;
   }
   m_buffer = new(nothrow) uint8_t[size];
   if (m_buffer == NULL)
   {
      error = CUSE_ERR_NO_MEMORY;
      return;
   }
   error = pthread_mutex_init(&m_mutex, NULL);
   if (!error)
   {
      m_state = CX88_DVB_BUF_MUTEX_CREATED;
      error = pthread_mutex_init(&m_read_lock, NULL);
   }
   if (!error)
   {
      m_state = CX88_DVB_BUF_READ_LOCK_CREATED;
      error = pthread_cond_init(&m_data_avail, NULL);
   }
   if (!error)
   {
      m_state = CX88_DVB_BUF_CV_CREATED;
   }
}

cx88_dvb_buffer::~cx88_dvb_buffer(void)
{
   delete[] m_buffer;
   m_buffer = NULL;
   m_size = 0;
   m_read_pos = 0;
   m_write_pos = 0;
   if (m_state >= CX88_DVB_BUF_CV_CREATED)
   {
      pthread_cond_destroy(&m_data_avail);
   }
   if (m_state >= CX88_DVB_BUF_READ_LOCK_CREATED)
   {
      pthread_mutex_destroy(&m_read_lock);
   }
   if (m_state >= CX88_DVB_BUF_MUTEX_CREATED)
   {
      pthread_mutex_destroy(&m_mutex);
   }
   m_state = CX88_DVB_BUF_UNINITIALIZED;
}

int cx88_dvb_buffer::read(void *user_ptr, int fflags, size_t &size)
{
   int error = 0;
   size_t bytes_remaining = size;
   pthread_mutex_lock(&m_read_lock);
   pthread_mutex_lock(&m_mutex);
   m_canceled = false;
   while (bytes_remaining > 0)
   {
      if (!m_write_rollover && (m_write_pos == m_read_pos))
      {
         if (fflags & CUSE_FFLAG_NONBLOCK)
         {
            break;
         }
         if (!m_timeout)
         {
            pthread_cond_wait(&m_data_avail, &m_mutex);
         }
         else
         {
            if (pthread_cond_timedwait(&m_data_avail, &m_mutex, &m_deadline) != 0)
            {
               error = CUSE_ERR_OTHER;
               break;
            }
         }
      }
      if (m_canceled)
      {
         error = CUSE_ERR_SIGNAL;
         break;
      }
      if (!m_write_rollover && (m_write_pos == m_read_pos))
      {
         if (cuse_got_peer_signal() == 0)
         {
            error = CUSE_ERR_SIGNAL;
            break;
         }
         else
         {
            continue;
         }
      }
      size_t write_pos = ((m_break_pos < m_size) ? m_break_pos : m_write_pos);
      size_t bytes_to_copy = write_pos - m_read_pos;
      if (bytes_to_copy == 0)
      {
         bytes_to_copy = m_size;
      }
      if (m_read_pos > write_pos)
      {
         bytes_to_copy = m_size - m_read_pos + write_pos;
      }
      if (bytes_to_copy > bytes_remaining)
      {
         bytes_to_copy = bytes_remaining;
      }
      size_t remainder = 0;
      if ((m_read_pos + bytes_to_copy) > m_size)
      {
         remainder = m_read_pos + bytes_to_copy - m_size;
      }
      uint8_t *read_pos = m_buffer + m_read_pos;
      pthread_mutex_unlock(&m_mutex);
      error = cuse_copy_out(read_pos, (uint8_t*)user_ptr + size - bytes_remaining, bytes_to_copy - remainder);
      bytes_remaining -= (bytes_to_copy - remainder);
      if (!error && (remainder > 0))
      {
         error = cuse_copy_out(m_buffer, (uint8_t*)user_ptr + size - bytes_remaining, remainder);
         bytes_remaining -= remainder;
      }
      pthread_mutex_lock(&m_mutex);
      if (error)
      {
         break;
      }
      m_write_rollover = false;
      m_read_pos += bytes_to_copy;
      if (m_read_pos >= m_size)
      {
         m_read_pos -= m_size;
      }
      if (m_read_pos == m_break_pos)
      {
         m_break_pos = m_size;
         break;
      }
   }
   pthread_mutex_unlock(&m_mutex);
   pthread_mutex_unlock(&m_read_lock);
   size = size - bytes_remaining;
   return error;
}

int cx88_dvb_buffer::write(const void *data, size_t size)
{
   pthread_mutex_lock(&m_mutex);
   size_t space_avail = m_read_pos - m_write_pos;
   if (m_write_pos >= m_read_pos)
   {
      space_avail = m_size - m_write_pos + m_read_pos;
   }
   if (m_write_rollover || (space_avail < size))
   {
      pthread_mutex_unlock(&m_mutex);
      return CUSE_ERR_OTHER;
   }
   size_t remainder = 0;
   if ((m_write_pos + size) > m_size)
   {
      remainder = m_write_pos + size - m_size;
   }
   uint8_t *write_pos = m_buffer + m_write_pos;
   pthread_mutex_unlock(&m_mutex);
   memcpy(write_pos, data, size - remainder);
   if (remainder > 0)
   {
      memcpy(m_buffer, (uint8_t*)data + size - remainder, remainder);
   }
   pthread_mutex_lock(&m_mutex);
   m_write_pos += size;
   if (m_write_pos >= m_size)
   {
      m_write_pos -= m_size;
   }
   m_write_rollover = (m_write_pos == m_read_pos);
   pthread_cond_signal(&m_data_avail);
   cuse_poll_wakeup();
   pthread_mutex_unlock(&m_mutex);
   return 0; 
}

void cx88_dvb_buffer::reset_pointers(void)
{
   m_read_pos = 0;
   m_write_pos = 0;
   m_write_rollover = false;
   m_break_pos = m_size;
}

int cx88_dvb_buffer::resize(size_t newsize)
{
   if (newsize == m_size)
   {
      return 0;
   }
   int error = 0;
   pthread_mutex_lock(&m_mutex);
   if (m_write_rollover || (m_read_pos != m_write_pos))
   {
      error = CUSE_ERR_BUSY;
   }
   else
   {
      uint8_t *newbuffer = new(nothrow) uint8_t[newsize];
      if (newbuffer == NULL)
      {
         error = CUSE_ERR_NO_MEMORY;
      }
      else
      {
         delete[] m_buffer;
         m_buffer = newbuffer;
         m_size = newsize;
         reset_pointers();
      }
   }
   pthread_mutex_unlock(&m_mutex);
   return error;
}

int cx88_dvb_buffer::poll(int events)
{
   int mask = 0;
   pthread_mutex_lock(&m_mutex);
   if (m_write_rollover || (m_read_pos != m_write_pos))
   {
      mask = events & CUSE_POLL_READ;
   }
   pthread_mutex_unlock(&m_mutex);
   return mask;
}

void cx88_dvb_buffer::reset(void)
{
   pthread_mutex_lock(&m_mutex);
   reset_pointers();
   pthread_mutex_unlock(&m_mutex);
}

void cx88_dvb_buffer::set_timeout(unsigned long timeout_s)
{
   pthread_mutex_lock(&m_mutex);
   if (timeout_s)
   {
      m_timeout = true;
      clock_gettime(CLOCK_MONOTONIC, &m_deadline);
      m_deadline.tv_sec += timeout_s;
   }
   else
   {
      m_timeout = false;
   }   
   pthread_mutex_unlock(&m_mutex);
}

void cx88_dvb_buffer::cancel_transfer(void)
{
   pthread_mutex_lock(&m_mutex);
   m_canceled = true;
   pthread_cond_signal(&m_data_avail);
   pthread_mutex_unlock(&m_mutex);
}
